@@ -1,12 +1,15 @@
 from enum import Enum
+from ast import parse
 
 from .type_inference import *
 from .rewrite_for import RewriteFor
 from .rewrite_tuple_assign import RewriteTupleAssign
 from .rewrite_augassign import RewriteAugAssign
-from .rewrite_dataclass import RewriteDataclasses
+from .rewrite_import_plutusdata import RewriteImportPlutusData
+from .rewrite_import_typing import RewriteImportTyping
+from .rewrite_import import RewriteImport
 
-from pluthon import pluthon_ast as plt
+import pluthon as plt
 from uplc.uplc_ast import BuiltInFun
 
 STATEMONAD = "s"
@@ -14,31 +17,62 @@ STATEMONAD = "s"
 
 BinOpMap = {
     Add: {
-        IntegerType: BuiltInFun.AddInteger,
-        ByteStringType: BuiltInFun.AppendByteString,
+        IntegerType: {
+            IntegerType: {
+                plt.AddInteger,
+            }
+        },
+        ByteStringType: {
+            ByteStringType: plt.AppendByteString,
+        },
     },
     Sub: {
-        IntegerType: BuiltInFun.SubtractInteger,
+        IntegerType: {
+            IntegerType: plt.SubtractInteger,
+        }
     },
     Mult: {
-        IntegerType: BuiltInFun.MultiplyInteger,
+        IntegerType: {
+            IntegerType: plt.MultiplyInteger,
+        }
     },
     Div: {
-        IntegerType: BuiltInFun.DivideInteger,
+        IntegerType: {
+            IntegerType: plt.DivideInteger,
+        }
     },
     Mod: {
-        IntegerType: BuiltInFun.RemainderInteger,
-    }
+        IntegerType: {
+            IntegerType: plt.ModInteger,
+        }
+    },
 }
 
 CmpMap = {
     Eq: {
-        IntegerType: BuiltInFun.EqualsInteger,
-        ByteStringType: BuiltInFun.EqualsByteString,
+        IntegerType: {
+            IntegerType: {
+                plt.EqualsInteger,
+            }
+        },
+        ByteStringType: {
+            ByteStringType: {
+                plt.EqualsByteString,
+            }
+        },
+        BoolType: {
+            BoolType: {
+                plt.EqualsBool,
+            }
+        },
     },
     Lt: {
-        IntegerType: BuiltInFun.LessThanInteger,
-    }
+        IntegerType: {
+            IntegerType: {
+                BuiltInFun.LessThanInteger,
+            }
+        },
+    },
 }
 
 ConstantMap = {
@@ -46,46 +80,30 @@ ConstantMap = {
     bytes: plt.ByteString,
     int: plt.Integer,
     bool: plt.Bool,
-    # TODO support higher level Optional type
-    type(None): plt.Unit,
+    type(None): lambda _: plt.NoneData(),
 }
 
-def extend_statemonad(names: typing.List[str], values: typing.List[plt.AST], old_statemonad: plt.UpdatableMap):
-    return plt.extend_map([n.encode() for n in names], values, old_statemonad)
+
+def extend_statemonad(
+    names: typing.List[str],
+    values: typing.List[plt.AST],
+    old_statemonad: plt.FunctionalMap,
+):
+    return plt.FunctionalMapExtend(old_statemonad, [n.encode() for n in names], values)
 
 
 class PythonBuiltIn(Enum):
     print = plt.Lambda(
-        ["f", "x"],
-        plt.Apply(
-            plt.Force(
-                    plt.BuiltIn(BuiltInFun.Trace),
-            ),
-            plt.Var("x"),
-            plt.Unit(),
-        )
+        ["f", "x", STATEMONAD],
+        plt.Trace(plt.Var("x"), plt.NoneData()),
     )
-    range = plt.Lambda(
-        ["f", "limit"],
-        plt.Tuple(
-            plt.Integer(0),
-            plt.Lambda(
-                ["f", "state"],
-                plt.Tuple(
-                    plt.Apply(plt.BuiltIn(BuiltInFun.LessThanInteger), plt.Var("state"), plt.Var("limit")),
-                    plt.Var("state"),
-                    plt.Apply(plt.BuiltIn(BuiltInFun.AddInteger), plt.Var("state"), plt.Integer(1)),
-                )
-            )
-        )
-    )
-    int = plt.Lambda(["f", "x"], plt.Apply(plt.BuiltIn(BuiltInFun.UnIData), plt.Var("x")))
-    __fields__ = plt.Lambda(["f", "x"],  plt.Apply(plt.Force(plt.Force(plt.BuiltIn(BuiltInFun.SndPair))), plt.Apply(plt.BuiltIn(BuiltInFun.UnConstrData), plt.Var("x"))))
+    range = plt.Lambda(["f", "limit", STATEMONAD], plt.Range(plt.Var("limit")))
+
 
 INITIAL_STATE = extend_statemonad(
     [b.name for b in PythonBuiltIn],
     [b.value for b in PythonBuiltIn],
-    plt.UpdatableMap(),
+    plt.FunctionalMap(),
 )
 
 
@@ -93,16 +111,16 @@ class UPLCCompiler(NodeTransformer):
     """
     Expects a TypedAST and returns UPLC/Pluto like code
     """
+
     def visit(self, node):
         """Visit a node."""
         node_class_name = node.__class__.__name__
         if node_class_name.startswith("Typed"):
-            node_class_name = node_class_name[len("Typed"):]
-        method = 'visit_' + node_class_name
+            node_class_name = node_class_name[len("Typed") :]
+        method = "visit_" + node_class_name
         visitor = getattr(self, method, self.generic_visit)
         return visitor(node)
 
-
     def visit_sequence(self, node_seq: typing.List[typedstmt]) -> plt.AST:
         s = plt.Var(STATEMONAD)
         for n in node_seq:
@@ -114,16 +132,22 @@ class UPLCCompiler(NodeTransformer):
         opmap = BinOpMap.get(type(node.op))
         if opmap is None:
             raise NotImplementedError(f"Operation {node.op} is not implemented")
-        op = opmap.get(node.typ)
-        if op is None:
-            raise NotImplementedError(f"Operation {node.op} is not implemented for type {node.typ}")
+        opmap2 = opmap.get(node.left.typ)
+        if opmap2 is None:
+            raise NotImplementedError(
+                f"Operation {node.op} is not implemented for left type {node.left.typ}"
+            )
+        op = opmap2.get(node.right.typ)
+        if opmap2 is None:
+            raise NotImplementedError(
+                f"Operation {node.op} is not implemented for left type {node.left.typ} and right type {node.right.typ}"
+            )
         return plt.Lambda(
             [STATEMONAD],
-            plt.Apply(
-                plt.BuiltIn(op),
+            op(
                 plt.Apply(self.visit(node.left), plt.Var(STATEMONAD)),
                 plt.Apply(self.visit(node.right), plt.Var(STATEMONAD)),
-            )
+            ),
         )
 
     def visit_Compare(self, node: Compare) -> plt.AST:
@@ -132,41 +156,61 @@ class UPLCCompiler(NodeTransformer):
         opmap = CmpMap.get(type(node.ops[0]))
         if opmap is None:
             raise NotImplementedError(f"Operation {node.ops[0]} is not implemented")
-        op = opmap.get(node.left.typ)
+        opmap2 = opmap.get(node.left.typ)
+        if opmap2 is None:
+            raise NotImplementedError(
+                f"Operation {node.ops[0]} is not implemented for left type {node.left.typ}"
+            )
+        op = opmap2.get(node.comparators[0].typ)
         if op is None:
-            raise NotImplementedError(f"Operation {node.ops[0]} is not implemented for type {node.left.typ}")
+            raise NotImplementedError(
+                f"Operation {node.ops[0]} is not implemented for left type {node.left.typ} and right type {node.comparators[0].typ}"
+            )
         return plt.Lambda(
             [STATEMONAD],
-            plt.Apply(
-                plt.BuiltIn(op),
+            op(
                 plt.Apply(self.visit(node.left), plt.Var(STATEMONAD)),
                 plt.Apply(self.visit(node.comparators[0]), plt.Var(STATEMONAD)),
-            )
+            ),
         )
-    
+
     def visit_Module(self, node: TypedModule) -> plt.AST:
         cp = plt.Program(
             "0.0.1",
-            plt.Let([(
-                "g",
-                plt.Apply(plt.Apply(self.visit_sequence(node.body), INITIAL_STATE), plt.ByteString("main".encode("utf8")))
-            )],
-            plt.Apply(plt.Var("g"), plt.Var("g")))
+            # TODO directly unwrap supposedly int/byte data? how?
+            plt.Let(
+                [
+                    (
+                        "g",
+                        plt.Apply(
+                            plt.Apply(self.visit_sequence(node.body), INITIAL_STATE),
+                            plt.ByteString("main".encode("utf8")),
+                        ),
+                    )
+                ],
+                plt.Apply(plt.Var("g"), plt.Var("g")),
+            ),
         )
         return cp
 
     def visit_Constant(self, node: TypedConstant) -> plt.AST:
         plt_type = ConstantMap.get(type(node.value))
         if plt_type is None:
-            raise NotImplementedError(f"Constants of type {type(node.value)} are not supported")
+            raise NotImplementedError(
+                f"Constants of type {type(node.value)} are not supported"
+            )
         return plt.Lambda([STATEMONAD], plt_type(node.value))
 
     def visit_NoneType(self, _: typing.Optional[typing.Any]) -> plt.AST:
         return plt.Lambda([STATEMONAD], plt.Unit())
 
     def visit_Assign(self, node: TypedAssign) -> plt.AST:
-        assert len(node.targets) == 1, "Assignments to more than one variable not supported yet"
-        assert isinstance(node.targets[0], Name), "Assignments to other things then names are not supported"
+        assert (
+            len(node.targets) == 1
+        ), "Assignments to more than one variable not supported yet"
+        assert isinstance(
+            node.targets[0], Name
+        ), "Assignments to other things then names are not supported"
         compiled_e = self.visit(node.value)
         # (\{STATEMONAD} -> (\x -> if (x ==b {self.visit(node.targets[0])}) then ({compiled_e} {STATEMONAD}) else ({STATEMONAD} x)))
         return plt.Lambda(
@@ -175,13 +219,20 @@ class UPLCCompiler(NodeTransformer):
                 [node.targets[0].id],
                 [plt.Apply(compiled_e, plt.Var(STATEMONAD))],
                 plt.Var(STATEMONAD),
-            )
+            ),
         )
 
     def visit_Name(self, node: TypedName) -> plt.AST:
         # depending on load or store context, return the value of the variable or its name
         if isinstance(node.ctx, Load):
-            return plt.Lambda([STATEMONAD], plt.Apply(plt.Var(STATEMONAD), plt.ByteString(node.id.encode())))
+            return plt.Lambda(
+                [STATEMONAD],
+                plt.FunctionalMapAccess(
+                    plt.Var(STATEMONAD),
+                    plt.ByteString(node.id.encode()),
+                    plt.Trace("NameError", plt.Error()),
+                ),
+            )
         raise NotImplementedError(f"Context {node.ctx} not supported")
 
     def visit_Expr(self, node: TypedExpr) -> plt.AST:
@@ -191,12 +242,9 @@ class UPLCCompiler(NodeTransformer):
         return plt.Lambda(
             [STATEMONAD],
             plt.Apply(
-                plt.Lambda(
-                    ["_"],
-                    plt.Var(STATEMONAD)
-                ),
+                plt.Lambda(["_"], plt.Var(STATEMONAD)),
                 plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
-            )
+            ),
         )
 
     def visit_Call(self, node: TypedCall) -> plt.AST:
@@ -210,15 +258,9 @@ class UPLCCompiler(NodeTransformer):
                     plt.Var("g"),
                     # pass the function to itself for recursion
                     plt.Var("g"),
-                    *(
-                        plt.Apply(
-                            self.visit(a),
-                            plt.Var(STATEMONAD)
-                        )
-                        for a in node.args
-                    )
-                )
-            )
+                    *(plt.Apply(self.visit(a), plt.Var(STATEMONAD)) for a in node.args),
+                ),
+            ),
         )
 
     def visit_FunctionDef(self, node: TypedFunctionDef) -> plt.AST:
@@ -226,7 +268,9 @@ class UPLCCompiler(NodeTransformer):
         if not isinstance(body[-1], Return):
             tr = Return(None)
             tr.typ = UnitType
-            assert node.typ.rettyp == UnitType, "Function has no return statement but is supposed to return not-None value"
+            assert (
+                node.typ.rettyp == UnitType
+            ), "Function has no return statement but is supposed to return not-None value"
             body.append(tr)
         compiled_body = self.visit_sequence(body[:-1])
         compiled_return = self.visit(body[-1].value)
@@ -248,14 +292,14 @@ class UPLCCompiler(NodeTransformer):
                             plt.Apply(
                                 compiled_body,
                                 args_state,
-                            )
-                        )
+                            ),
+                        ),
                     )
                 ],
                 plt.Var(STATEMONAD),
-            )
+            ),
         )
-    
+
     def visit_While(self, node: TypedWhile) -> plt.AST:
         compiled_c = self.visit(node.test)
         compiled_s = self.visit_sequence(node.body)
@@ -275,20 +319,26 @@ class UPLCCompiler(NodeTransformer):
                             ["s", "f"],
                             plt.Ite(
                                 plt.Apply(compiled_c, plt.Var("s")),
-                                plt.Apply(plt.Var("f"), plt.Apply(compiled_s, plt.Var("s")), plt.Var("f")),
+                                plt.Apply(
+                                    plt.Var("f"),
+                                    plt.Apply(compiled_s, plt.Var("s")),
+                                    plt.Var("f"),
+                                ),
                                 plt.Var("s"),
-                            )
-                        )
+                            ),
+                        ),
                     ),
                 ],
-                term=plt.Apply(plt.Var("g"), plt.Var(STATEMONAD), plt.Var("g"))
-            )
+                term=plt.Apply(plt.Var("g"), plt.Var(STATEMONAD), plt.Var("g")),
+            ),
         )
 
     def visit_For(self, node: TypedFor) -> plt.AST:
         # TODO implement for list
         if isinstance(node.iter.typ, ListType):
-            raise NotImplementedError("Compilation of list iterators not implemented yet.")
+            raise NotImplementedError(
+                "Compilation of list iterators not implemented yet."
+            )
         raise NotImplementedError("Compilation of raw for statements not supported")
 
     def visit_If(self, node: TypedIf) -> plt.AST:
@@ -298,80 +348,135 @@ class UPLCCompiler(NodeTransformer):
                 plt.Apply(self.visit(node.test), plt.Var(STATEMONAD)),
                 plt.Apply(self.visit_sequence(node.body), plt.Var(STATEMONAD)),
                 plt.Apply(self.visit_sequence(node.orelse), plt.Var(STATEMONAD)),
-            )
+            ),
+        )
+
+    def visit_Return(self, node: TypedReturn) -> plt.AST:
+        raise NotImplementedError(
+            "Compilation of return statements except for last statement in function is not supported."
         )
-    
-    def visit_Return(self, node: TypedReturn) -> plt.AST:
-        raise NotImplementedError("Compilation of return statements except for last statement in function is not supported.")
 
-    
     def visit_Pass(self, node: TypedPass) -> plt.AST:
         return self.visit_sequence([])
 
     def visit_Subscript(self, node: TypedSubscript) -> plt.AST:
-        assert isinstance(node.slice, Index), "Only single index slices are currently supported"
+        assert isinstance(
+            node.slice, Index
+        ), "Only single index slices are currently supported"
         if isinstance(node.value.typ, TupleType):
-            assert isinstance(node.slice.value, Constant), "Only constant index access for tuples is supported"
+            assert isinstance(
+                node.slice.value, Constant
+            ), "Only constant index access for tuples is supported"
             assert isinstance(node.ctx, Load), "Tuples are read-only"
-            return plt.Lambda([STATEMONAD], emulate_nth(
-                plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
-                node.slice.value.value,
-                len(node.value.typ.typs),
-            ))
+            return plt.Lambda(
+                [STATEMONAD],
+                plt.FunctionalTupleAccess(
+                    plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
+                    node.slice.value.value,
+                    len(node.value.typ.typs),
+                ),
+            )
         if isinstance(node.value.typ, ListType):
             # rf'(\{STATEMONAD} -> let g = (\i xs f -> if (!NullList xs) then (!Trace "OOB" (Error ())) else (if i ==i 0 then (!HeadList xs) else f (i -i 1) (!TailList xs) f)) in (g ({compiled_i} {STATEMONAD}) ({compiled_v} {STATEMONAD}) g))'
             return plt.Lambda(
                 [STATEMONAD],
                 plt.Let(
                     [
-                        ("g",
-                        plt.Lambda(["i", "xs", "f"],
-                           plt.Ite(
-                               plt.Force(plt.Apply(plt.BuiltIn(BuiltInFun.NullList), plt.Var("xs"))),
-                               self.visit(TypedCall(TypedName("print", ctx=Load()), [TypedConstant("Out of bounds", InstanceType("str"))])),
-                               plt.Ite(
-                                   plt.Apply(plt.BuiltIn(BuiltInFun.EqualsInteger), plt.Var("i"), plt.Integer(0)),
-                                   plt.Force(plt.Apply(plt.BuiltIn(BuiltInFun.HeadList), plt.Var("xs"))),
-                                   plt.Apply(
-                                       plt.Var("f"),
-                                       plt.Apply(plt.BuiltIn(BuiltInFun.SubtractInteger), plt.Var("i"), plt.Integer(1)),
-                                       plt.Force(plt.Apply(plt.BuiltIn(BuiltInFun.TailList), plt.Var("xs"))),
-                                       plt.Var("f"),
-                                   )
-                               )
-                           )
-                       ))
+                        (
+                            "g",
+                            plt.Lambda(
+                                ["i", "xs", "f"],
+                                plt.Ite(
+                                    plt.Force(
+                                        plt.Apply(
+                                            plt.BuiltIn(BuiltInFun.NullList),
+                                            plt.Var("xs"),
+                                        )
+                                    ),
+                                    self.visit(
+                                        TypedCall(
+                                            TypedName("print", ctx=Load()),
+                                            [
+                                                TypedConstant(
+                                                    "Out of bounds", InstanceType("str")
+                                                )
+                                            ],
+                                        )
+                                    ),
+                                    plt.Ite(
+                                        plt.Apply(
+                                            plt.BuiltIn(BuiltInFun.EqualsInteger),
+                                            plt.Var("i"),
+                                            plt.Integer(0),
+                                        ),
+                                        plt.Force(
+                                            plt.Apply(
+                                                plt.BuiltIn(BuiltInFun.HeadList),
+                                                plt.Var("xs"),
+                                            )
+                                        ),
+                                        plt.Apply(
+                                            plt.Var("f"),
+                                            plt.Apply(
+                                                plt.BuiltIn(BuiltInFun.SubtractInteger),
+                                                plt.Var("i"),
+                                                plt.Integer(1),
+                                            ),
+                                            plt.Force(
+                                                plt.Apply(
+                                                    plt.BuiltIn(BuiltInFun.TailList),
+                                                    plt.Var("xs"),
+                                                )
+                                            ),
+                                            plt.Var("f"),
+                                        ),
+                                    ),
+                                ),
+                            ),
+                        )
                     ],
                     plt.Apply(
                         plt.Var("g"),
                         plt.Apply(self.visit(node.slice.value), plt.Var(STATEMONAD)),
                         plt.Apply(self.visit(node.value), plt.Var(STATEMONAD)),
-                        plt.Var("g")
-                    )
-                )
-              )
+                        plt.Var("g"),
+                    ),
+                ),
+            )
         raise NotImplementedError(f"Could not implement subscript of {node}")
-    
+
     def visit_Tuple(self, node: TypedTuple) -> plt.AST:
         return plt.Lambda(
             [STATEMONAD],
-            plt.Tuple(
+            plt.FunctionalTuple(
                 *(plt.Apply(self.visit(e), plt.Var(STATEMONAD)) for e in node.elts)
-            )
+            ),
         )
 
     def visit_ClassDef(self, node: ClassDef) -> plt.AST:
+        # TODO add initializer to state monad
         return self.visit_sequence([])
 
     def visit_Attribute(self, node: TypedAttribute) -> plt.AST:
         # rewrite to access the field at position node.pos
         # (use the internal function for fields)
-        return self.visit(TypedSubscript(
-            value=TypedCall(TypedName(id="__fields__", ctx=Load(),typ=FunctionType([InstanceType], ListType([]))), [node.value], typ=ListType([])),
-            slice=Index(value=TypedConstant(value=node.pos, typ=InstanceType("int"))),
-            typ=node.typ,
-        ))
-
+        return self.visit(
+            TypedSubscript(
+                value=TypedCall(
+                    TypedName(
+                        id="__fields__",
+                        ctx=Load(),
+                        typ=FunctionType([InstanceType], ListType([])),
+                    ),
+                    [node.value],
+                    typ=ListType([]),
+                ),
+                slice=Index(
+                    value=TypedConstant(value=node.pos, typ=InstanceType("int"))
+                ),
+                typ=node.typ,
+            )
+        )
 
     def generic_visit(self, node: AST) -> plt.AST:
         raise NotImplementedError(f"Can not compile {node}")
@@ -379,13 +484,17 @@ class UPLCCompiler(NodeTransformer):
 
 def compile(prog: AST):
     compiler_steps = [
+        # Important to call this one first - it imports all further files
+        RewriteImport,
+        # The remaining order is almost arbitrary
         RewriteAugAssign,
         RewriteFor,
         RewriteTupleAssign,
-        RewriteDataclasses,
+        RewriteImportPlutusData,
+        RewriteImportTyping,
         AggressiveTypeInferencer,
-        UPLCCompiler
+        UPLCCompiler,
     ]
     for s in compiler_steps:
         prog = s().visit(prog)
-    return prog
\ No newline at end of file
+    return prog
